"""
This module contains functions for reducing detector energy binning of raw NEXAFS data from beamlines.

For instance, as of Apr 2024, the MEX2 beamline at the Australian Synchrotron has 4 detectors,
each with 4096 energy bins (channels), all of which is recorded as the beam energy is scanned.
"""

import numpy as np
import numpy.typing as npt
import os
from pyNexafs.utils.mda import MDAFileReader
import matplotlib.pyplot as plt
from typing import Callable, Literal


class reducer:
    """
    A class for reducing detector energy binning of raw NEXAFS data from beamlines.

    Parameters
    ----------
    energies : array_like
        A 1D array of energy values for the scanning beam energy.
    dataset : array_like
        The 2D/3D NEXAFS dataset with dimensions (beam_energy, detection_energy_bin, detector_index).
        The detector_index dimension is optional, used for multiple multi-channel analysers.
    bin_energies : array_like | None, optional
        A 1D array of energy values for each detection energy bin. If not provided, the bin numbers are used.
        If 2D, the second dimension must match the number of detectors.
    """

    def __init__(
        self,
        energies: npt.NDArray,
        dataset: npt.NDArray,
        bin_energies: npt.NDArray | None = None,
    ) -> None:
        self.set_data(energies, dataset, bin_energies)

    def set_data(
        self,
        energies: npt.NDArray,
        dataset: npt.NDArray,
        bin_energies: npt.NDArray | None = None,
    ) -> None:
        """
        Set the energies, dataset and bin_energies attributes.

        Parameters
        ----------
        energies : npt.NDArray
            A 1D array of energy values for the scanning beam energy.
        dataset : npt.NDArray
            The 3D NEXAFS dataset with dimensions (beam_energy, detection_energy_bin, detector_index).
            The detector_index dimension is optional, used for multiple multi-channel analysers.
        bin_energies : npt.NDArray | None, optional
            A 1D array of energy values for each detection energy bin. If not provided, the bin numbers are used.
            If 2D, the second dimension must match the number of detectors.
        """
        # Convert to numpy arrays
        energies = np.asarray(energies)
        dataset = np.asarray(dataset)
        if bin_energies is not None:
            bin_energies = np.asarray(bin_energies)

        # Check validity of input
        self._validify_inputs(energies, dataset, bin_energies)

        # Generate bin numbers if not provided
        if bin_energies is None:
            # Create bin numbers:
            bin_energies = np.linspace(0, dataset.shape[1], dataset.shape[1])
            self._bins_assigned = False
        else:
            self._bins_assigned = True

        # Assign attributes
        self._energies: np.ndarray = energies
        self._dataset: np.ndarray = dataset
        self._bin_energies: np.ndarray = bin_energies

        # New boolean attribute to track if bin_energies has been set
        self._bins_assigned: bool
        """Tracks if bin_energies has been set."""

    @property
    def energies(self) -> npt.NDArray:
        """
        The scanning beam energy values.

        Parameters
        ----------
        energies : npt.NDArray
            A 1D array of energy values for the scanning beam energy.
            Must have the same length as the first dimension of the dataset.

        Returns
        -------
        npt.NDArray
            The beam energy values.
        """
        return self._energies

    @energies.setter
    def energies(self, energies: npt.NDArray) -> None:
        # Raises an error if input is invalid.
        if reducer._validify_inputs(energies, self._dataset, self._bin_energies):
            self._energies = np.asarray(energies)

    @property
    def dataset(self) -> npt.NDArray:
        """
        The NEXAFS dataset with dimensions (beam_energy, detection_energy_bin, detector_index).

        Parameters
        ----------
        dataset : npt.NDArray
            The NEXAFS dataset indexed with dimensions (beam_energy, detection_energy_bin, detector_index).
            The detector_index dimension is optional, used for multiple multi-channel analysers.

        Returns
        -------
        npt.NDArray
            The NEXAFS dataset.
        """
        return self._dataset

    @dataset.setter
    def dataset(self, dataset: npt.NDArray) -> None:
        # Raises an error if input is invalid.
        if reducer._validify_inputs(self._energies, dataset, self._bin_energies):
            self._dataset = np.asarray(dataset)

    @property
    def bin_energies(self) -> npt.NDArray:
        """
        The energy values (or bin indexes) for each detection energy bin.

        Parameters
        ----------
        bin_energies : npt.NDArray | None
            A 1D or 2D array of energy values for each detection energy bin.
            Must have the same length as the second dimension of the `dataset`.
            If None, generates bin numbers indexed to `dataset`.

        Returns
        -------
        npt.NDArray
            The energy values (or bin indexes) for each detection energy bin.
            Can be 1D or 2D, matching the latter dimensions of the dataset.
        """
        return self._bin_energies

    @bin_energies.setter
    def bin_energies(self, bin_energies: npt.NDArray | None) -> None:
        if bin_energies is None:
            self._bins_assigned = False
            self._bin_energies = np.linspace(
                0, self.dataset.shape[1], self.dataset.shape[1]
            )
        else:
            # Raises an error if input is invalid.
            if reducer._validify_inputs(self._energies, self._dataset, bin_energies):
                self._bin_energies = np.asarray(bin_energies)

    @property
    def has_bin_energies(self) -> bool:
        """
        A boolean property indicating whether the bin_energies attribute has been set.

        Important in determining binning behaviour.

        Returns
        -------
        bool
            True if bin_energies has been set, otherwise False.
        """
        return self._bins_assigned

    @property
    def detectors(self) -> int:
        """
        The number of detectors in the dataset.

        Returns
        -------
        int
            The number of detectors in the dataset.
        """
        if self._dataset.ndim == 2:
            return 1
        else:
            return self._dataset.shape[2]

    def domain_to_detector_bin_index_range(
        self, bin_domain: tuple[int, int] | tuple[float, float], detector_idx: int = 0
    ) -> tuple[int, int]:
        """
        Convert an energy|bin domain to bin indexes for a given detector.

        By default, the first detector is assumed.
        If `bin_energies` are set, the index of the closest energy value to the domain is used.
        If no `bin_energies` are set, the domain is converted to the nearest integer and applied directly.

        Parameters
        ----------
        bin_domain : tuple[int, int]
            The energy domain to convert to bin indexes.
        detector_idx : int, optional
            Index of the detector to use, by default 0.

        Returns
        -------
        tuple[int, int]
            The lower and upper selected bin indexes for the given domain.
        """
        if not self.has_bin_energies:
            # If no labels, the range matches the indexes, convert to nearest integer
            return tuple(np.round(bin_domain).astype(int))
        else:
            if not isinstance(detector_idx, int):
                raise ValueError("`detector_idx` must be an integer.")
            # If labels and a given domain, convert the range to indexes
            if detector_idx >= self.detectors:
                raise ValueError(
                    f"Detector index {detector_idx} out of range [0-{self.detectors})."
                )
            else:
                if self.bin_energies.ndim == 1:  # Singular bin_energies
                    lb = np.argmin(np.abs(self.bin_energies - bin_domain[0]))
                    ub = np.argmin(np.abs(self.bin_energies - bin_domain[1]))
                else:  # Multiple bin_energies for detectors
                    lb = np.argmin(
                        np.abs(self.bin_energies[:, detector_idx] - bin_domain[0])
                    )
                    ub = np.argmin(
                        np.abs(self.bin_energies[:, detector_idx] - bin_domain[1])
                    )
                return (lb, ub)

    def domain_to_index_range(
        self,
        bin_domain: (
            list[tuple[int, int] | tuple[float, float]]
            | tuple[int, int]
            | tuple[float, float]
            | None
        ),
    ) -> list[tuple[int, int]]:
        """
        Convert an energy|bin domain to bin indexes for all detectors.

        Uses `domain_to_detector_bin_indexes` to convert the domain for each detector.

        Parameters
        ----------
        bin_domain : list[tuple[int, int] | tuple[float, float]] | tuple[int, int] | tuple[float, float]
            The energy|bin domain to convert to bin indexes.

            If None, indexes of the full dataset is returned.

        Returns
        -------
        list[tuple[int, int]]
            The lower and upper selected bin indexes for each detector.
        """
        if bin_domain is None:
            # Generate the full range of indexes
            return [(0, self.bin_energies.shape[0])] * self.detectors
        elif isinstance(bin_domain, list):
            return [
                self.domain_to_detector_bin_index_range(
                    bin_domain=domain, detector_idx=i
                )
                for i, domain in enumerate(bin_domain)
            ]
        else:
            if not self.has_bin_energies:
                # Can repeat the domain for each detector; indexing is the same
                return [
                    self.domain_to_detector_bin_index_range(bin_domain=bin_domain)
                ] * self.detectors
            else:
                return [
                    self.domain_to_detector_bin_index_range(
                        bin_domain=bin_domain, detector_idx=i
                    )
                    for i in range(self.detectors)
                ]

    def reduce_domain(
        self,
        bin_domain: (
            list[tuple[float, float] | tuple[int, int]]
            | tuple[float, float]
            | tuple[int, int]
            | None
        ) = None,
    ) -> tuple[npt.NDArray, npt.NDArray]:
        """
        Reduce the dataset to a specified energy bin domain.

        Parameters
        ----------
        bin_domain : list[tuple[float, float] | tuple[int, int]] | tuple[float, float] | tuple[int, int] | None, optional
            The domain to reduce the dataset to. If None, the dataset is not reduced.
            If a single tuple is provided, the domain is applied to all detectors.
            If a list of tuples is provided, the domain is applied to each detector.
            The tuples should be in the form (min, max) for each detector.
            By default `bin_domain` is treated as a set of energy values.
            If has_bin_energies is False, the domain is applied to the bin indexes instead.

        Returns
        -------
        reduced_bin_energies : npt.NDArray
            The reduced energy values for each detection energy bin.
            If bin_energies is not set, the values correspond bin indexes.
            Like `reduced_dataset`, `reduced_bin_energies` has as dimensions (N, D).
        reduced_dataset : npt.NDArray
            The reduced NEXAFS dataset, with dimensions (M, N, D),
            where M is the number of beam energies,
            N is the reduced number of detection energy bins according to the provided domain,
            and D is the number of detectors.
        """
        # Prepare the domain variable
        if bin_domain is None:
            # Return the full dataset window if no domain
            return self.bin_energies.copy(), self.dataset.copy()
        else:
            # Check if list of int/float, without tuples and length 2, then convert to tuple.
            if (
                isinstance(bin_domain, list)
                and len(bin_domain) == 2
                and all(isinstance(i, (int, float)) for i in bin_domain)
            ):
                bin_domain = tuple(bin_domain)

            # Convert non-list to list
            if not isinstance(bin_domain, list):
                bin_domain = [bin_domain] * self.detectors
            # Check list validity
            if len(bin_domain) != self.detectors:
                raise ValueError(
                    f"`domain` must be singular i.e. `(min, max)`, or have the same length ({len(bin_domain)}) as"
                    + f" the number of detectors ({self.detectors}) i.e. `[(min1, max1), ..., (minN, maxN)]`."
                )

            # Reduce the dataset to the domain and a numpy array
            reduced_idxs = np.asarray(self.domain_to_index_range(bin_domain))
            # always 2D ^

            if reduced_idxs.ndim == 2 and self.dataset.ndim == 2:
                if reduced_idxs.shape[0] == 1:
                    lb, ub = reduced_idxs[0]
                    if self.has_bin_energies:
                        # Same logic for single or multiple detectors:
                        return self.bin_energies[lb:ub], self.dataset[:, lb:ub]
                    else:
                        return np.arange(lb, ub), self.dataset[:, lb:ub]
                else:
                    raise ValueError(
                        f"Cannot reduce a single detector dataset (shape: {self.dataset.shape}) with multiple detector domains {bin_domain}."
                    )
            elif reduced_idxs.ndim == 2 and self.dataset.ndim == 3:
                data = []
                bins = []
                for i, (lb, ub) in enumerate(reduced_idxs):
                    # Add the slice of the dataset
                    data.append(self.dataset[:, lb:ub, i])
                    # Add the slice of the bin_energies
                    if self.has_bin_energies:
                        if self.bin_energies.ndim == 1:
                            bins.append(self.bin_energies[lb:ub])
                        else:
                            bins.append(self.bin_energies[lb:ub, i])
                    else:
                        bins.append(np.arange(lb, ub))
                # Convert to arrays
                data = np.array(data)
                bins = np.array(bins)
                # Swap back the indexes of the datasets (currently detector, energies, bins):
                data = np.moveaxis(data, 0, -1)
                bins = np.moveaxis(bins, 0, -1) if bins.ndim == 2 else bins
                return bins, data
            else:
                raise ValueError(
                    f"Invalid domain shape {reduced_idxs.shape} for dataset shape {self.dataset.shape}."
                )

    def reduce(
        self,
        fn: Callable[[npt.NDArray, npt.NDArray, npt.NDArray], npt.NDArray],
        bin_domain: (
            list[tuple[float, float] | tuple[int, int]]
            | tuple[float, float]
            | tuple[int, int]
            | None
        ) = None,
    ) -> tuple[npt.NDArray, npt.NDArray]:
        """
        Reduce the bin dimension of the dataset to a singular value.

        Parameters
        ----------
        fn : Callable[[npt.NDArray, npt.NDArray, npt.NDArray], npt.NDArray]
            The function to apply to the reducer data.
            Parameters (in order) are energies, dataset, bin_energies.
            See examples for usage.

        bin_domain : list[tuple[float, float] | tuple[int, int]] | tuple[float, float] | tuple[int, int] | None, optional
            The energy/index domain to reduce the dataset energy bins to. If None, the dataset is not reduced.
            If a single tuple is provided, the domain is applied to all detectors.
            If a list of tuples is provided, the domain is applied to each detector.
            The tuples should be in the form (min, max) for each detector.

        Returns
        -------
        bin_energies : npt.NDArray
            An array of the bin energies within the bin domain.
            If bin_energies is not set, the values correspond bin indexes.
            Can be 1D if one detector, or 2D for multiple detectors.
        reduced_dataset : npt.NDArray
            The reduced NEXAFS dataset.
            Shape depends on the return of `fn`.

        Examples
        --------
        # Example 1: Sum the dataset over the bin and detector dimensions
        def sum_dataset(energies, dataset, bin_energies):
            return dataset.sum(axis=(1, 2))
        dataset = reducer.reduce_bins_to_singular(sum_dataset)

        # Example 2: Translate the dataset to be positive definite,
        # and add 0.01 so that the log is not undefined.
        def positive_definite(energies, dataset, bin_energies):
            return np.log(dataset - dataset.min() + 1e-2)
        """
        # Check the validity of the fn input.
        if not callable(fn):
            raise ValueError("`fn` must be a callable function.")

        # Collect possible inputs for the function
        kw_pairs = {
            "energies": self.energies,
            "dataset": self.dataset,
            "bin_energies": self.bin_energies,
        }

        # Check that the function signature contains the correct variable names,
        # and collect the correct inputs for the function.
        kwargs = {}
        for kw in fn.__code__.co_varnames:
            if kw not in kw_pairs:
                raise ValueError(
                    f"`fn` {fn} has an invalid keyword: {kw}."
                    + "Only allowed parameters are `energies`, `dataset`, and `bin_energies`."
                )
            else:
                # Add to kwargs
                kwargs[kw] = kw_pairs[kw]

        if bin_domain is not None:
            reduced_bin_energies, kwargs["dataset"] = self.reduce_domain(bin_domain)
        else:
            reduced_bin_energies, kwargs["dataset"] = (
                self.bin_energies.copy(),
                self.dataset.copy(),
            )
        # print(reduced_bin_energies)

        # Apply the function to the sub_domain dataset
        reduced_dset = fn(**kwargs)
        # Return the reduced dataset. Don't need to return bin_energies as
        return reduced_bin_energies, reduced_dset

    @staticmethod
    def _sum_fn_factor(sum_axes: tuple[int, ...]) -> Callable:
        def fn(
            energies: npt.NDArray, dataset: npt.NDArray, bin_energies: npt.NDArray
        ) -> npt.NDArray:
            return dataset.sum(axis=sum_axes)

        return fn

    def reduce_by_sum(
        self,
        bin_domain: (
            list[tuple[float, float] | tuple[int, int]]
            | tuple[float, float]
            | tuple[int, int]
            | None
        ) = None,
        axis: Literal["bin_energies", "detectors"] | None = None,
    ) -> tuple[npt.NDArray, npt.NDArray]:
        """
        Reduce the dataset into a single value for each beam energy.

        The single value is created by summing over all dimensions (bin_energies, detectors),
        unless dimensions are specified by `axis`, in which case the sum is only over the specified axes.

        Parameters
        ----------
        bin_domain : list[tuple[float, float] | tuple[int, int]] | tuple[float, float] | tuple[int, int] | None, optional
            The energy/index domain to reduce the dataset to. If None, the dataset is not reduced.
            If a single tuple is provided, the domain is applied to all detectors.
            If a list of tuples is provided, the domain is applied to each detector.
            The tuples should be in the form (min, max) for each detector.
        axis : Literal["bin_energies", "detectors"] | None, optional
            The axis or axes to sum over. If None, the dataset is summed over all axes.

        Returns
        -------
        reduced_bin_energies : npt.NDArray
            The domain-reduced energy values for each detection energy bin.
            If `bin_energies` is not set, the values correspond bin indexes.
            The dimensions are (N, D), where N is the reduced number of
            detection energy bins, and D is the number of detectors.
        reduced_dataset : npt.NDArray
            The reduced NEXAFS dataset, with dimensions (M)
            where M is the number of incident beam energies.
        """
        if axis is not None and axis not in ["bin_energies", "detectors"]:
            raise ValueError("`axis` must be either 'bin_energies' or 'detectors'.")
        if axis == "bin_energies":
            # f = lambda energies, dataset, bin_energies: dataset.sum(axis=1)
            f = self._sum_fn_factor(sum_axes=(1,))
        elif axis == "detectors":
            f = self._sum_fn_factor(sum_axes=(2,))
        else:
            f = self._sum_fn_factor(sum_axes=(1, 2))
        return self.reduce(fn=f, bin_domain=bin_domain)

    def reduce_to_bin_features(self) -> tuple[npt.NDArray, npt.NDArray]:
        """
        Reduce the dataset to find the signal dependence of energy bins with signal.

        The return signal has
        1. Summed over all beam energies.
        2. Translated to be positive definite (a base 1e-2 is added so we can log plot).
        the sum of the dataset over all beam energies.

        Returns
        -------
        bin_energies : npt.NDArray
            The energy values for each detection energy bin.
            If bin_energies is not set, the values correspond bin indexes.
        dataset_sum : npt.NDArray
            The reduced dataset, with dimensions (N, D),
            where N is the number of detection energy bins with signal,
            and D is the number of detectors.
        """
        # Sum over all energies to find the energy bins with signal.
        ds_sum: np.ndarray = self.dataset.sum(axis=0)
        # Translate every bin to be positive definite.
        if np.any(ds_sum <= 0):
            # Keep relative signal amplitude between detectors
            ds_sum += -ds_sum.min()
            # Add a base 1e-2 so we can log plot
            ds_sum += 1e-2

        return self.bin_energies, ds_sum

    @staticmethod
    def _validify_inputs(
        energies: npt.NDArray,
        dataset: npt.NDArray,
        bin_energies: npt.NDArray | None = None,
    ) -> bool:
        """
        Validate the input dataset, energies and bin_energies.

        Raises a ValueError if the inputs are invalid, otherwise returns True.

        Parameters
        ----------
        energies : npt.NDArray
            A 1D array of energy values for the scanning beam energy.
        dataset : npt.NDArray
            The NEXAFS dataset with dimensions (beam_energy, detection_energy_bin, detector_index).
            The detector_index dimension is optional, used for multiple multi-channel analysers.
        bin_energies : npt.NDArray, optional
            A 1D array of energy values for each detection energy bin. If not provided, the bin numbers are used.
            If 2D, the second dimension must match the number of detectors.

        Returns
        -------
        tuple[npt.NDArray, npt.NDArray, npt.NDArray]
            The validated dataset, energies and bin_energies.
        """
        if not energies.ndim == 1:
            raise ValueError("`energies` must be a 1D numpy array.")
        if not (dataset.ndim == 2 or dataset.ndim == 3):
            raise ValueError(
                "`dataset` must be a 2D or 3D numpy array with indexes corresponding to"
                + " (beam_energy, detection_energy_bin, detector_index). The detector_index dimension is optional."
            )
        if not (energies.shape[0] == dataset.shape[0]):
            raise ValueError(
                f"`energies` must have the same length ({energies.shape[0]})"
                + f" as the first dimension of `dataset` ({dataset.shape[0]})."
            )
        if dataset.ndim == 2:
            dataset = dataset[:, :, np.newaxis]  # add final dimension.

        # Check bin_energies validity
        if bin_energies is not None:
            bin_energies = np.asarray(bin_energies)
            if not (bin_energies.ndim == 1 or bin_energies.ndim == 2):
                raise ValueError(
                    "`bin_energies` must be a 1D or 2D numpy array, with indexes corresponding to"
                    + " (detection_bin_energy, detector_index). The second dimension is optional."
                )
            if not (bin_energies.shape[0] == dataset.shape[1]):
                raise ValueError(
                    f"`bin_energies` must have the same length ({bin_energies.shape[0]}) "
                    + f"as the second dimension of `dataset` ({dataset.shape[1]})."
                )
            if bin_energies.ndim == 2 and not (
                bin_energies.shape[1] == dataset.shape[2]
            ):
                raise ValueError(
                    f"`bin_energies` must have the same length ({bin_energies.shape[1]}) "
                    + f"as the third dimension of `dataset` ({dataset.shape[2]})."
                )
        return True


if __name__ == "__main__":
    ###
    ### This is a demo to show how to use the reducer class.
    ###

    path = os.path.dirname(__file__)
    package_path = os.path.normpath(os.path.join(path, "../../"))
    mda_path = os.path.normpath(
        os.path.join(package_path, "tests/test_data/au/MEX2/MEX2_5643.mda")
    )
    mda_data, mda_scans = MDAFileReader(mda_path).read_scans()
    data1D, data2D = mda_data

    # Convert MEX2 Data
    energies = data1D[:, 0] * 1000  # Convert keV to eV
    span = [96, 500]
    binned_data = data2D[:, span[0] : span[1], :]
    TOTAL_BINS = span[1] - span[0]
    BIN_ENERGY = 11.935
    BIN_96_ENERGY = 1146.7
    bin_energies = np.linspace(
        BIN_96_ENERGY, BIN_96_ENERGY + TOTAL_BINS * BIN_ENERGY, TOTAL_BINS
    )

    # Setup a graph
    graph = plt.subplots(1, 3, figsize=(12, 6))
    fig: plt.Figure = graph[0]
    ax: list[plt.Axes] = graph[1]

    # Manually reduce
    reduced_data = binned_data.sum(axis=1)  # Sum over bins
    [
        ax[0].plot(energies, reduced_data[:, i], label=f"Detector {i}")
        for i in range(reduced_data.shape[1])
    ]
    ax[0].set_title("Raw data, summed over all bin energies")
    ax[0].set_xlabel("Beam Energy (eV)")
    ax[0].set_ylabel("Intensity")
    ax[0].legend()

    # Now use the reducer
    red = reducer(energies, binned_data, bin_energies)
    bin_features, dataset_sum = red.reduce_to_bin_features()
    ax[1].plot(bin_features, dataset_sum, label=[f"Detector {i}" for i in [1, 2, 3, 4]])
    ax[1].set_title("Raw data, summed over all beam energies")
    ax[1].set_yscale("log")
    # ax[1].set_ylim(1e-2, logdata.max()*5)
    ax[1].set_xlabel("Bin Energy (eV)")
    ax[1].set_ylabel("Intensity")
    ax[1].legend()

    # Demonstrate bin domain use
    domain = (3300, 3700)
    # Demonstrate bin reduction
    reduced_bin_energies, reduced_dset = red.reduce_by_sum(bin_domain=domain)
    ax[2].plot(energies, reduced_dset, label="Reduced")
    ax[2].legend()
    ax[2].set_title("Reduced NEXAFS data")
    ax[2].set_xlabel("Beam Energy (eV)")
    ax[2].set_ylabel("Intensity")

    fig.tight_layout()
    plt.show()



"""
Module to handle default normalisation settings to be used across datasets.
"""

from enum import Enum, EnumType
from abc import ABC, abstractmethod, ABCMeta
from io import TextIOWrapper
import yaml
import os


class configMeta(ABCMeta):
    """
    Metaclass for all normalisation settings classes.

    Ensures every subclass has default arg values, so that configurations can
    be initialised and set.

    Raises
    ------
    NotImplementedError
        If the subclass does not have default arguments for __init__ variables.
    """

    def __new__(cls: type, name: str, bases: tuple[type, ...], dct: dict) -> type:
        """
        The metaclass' class creator.

        Check that each constructor argument has a default value.

        Parameters
        ----------
        cls : type
            The metaclass type.
        name : str
            The name of the class being created.
        bases : tuple[type]
            The base classes of the class being created.
        dct : dict
            The class attributes and methods dictionary.

        Returns
        -------
        type
            The new class type with the metaclass applied.
        """
        # Check the arguments for the class constructor and subclasses
        init_fns = []
        # The the class init
        if "__init__" in dct:
            init_fns.append(dct["__init__"])
        # Get subclass inits
        for key, val in dct.items():
            if hasattr(val, "__init__") and key != "__init__" and isinstance(val, type):
                if issubclass(val, cls):  # Was collecting enum attribute inits.
                    init_fns.append(val.__init__)

        # Check all inits for default arguments to allow instantiation
        for init_fn in init_fns:
            args: tuple = init_fn.__code__.co_varnames[1:]
            defs: list = (
                init_fn.__defaults__ if init_fn.__defaults__ is not None else []
            )
            if "args" in args:
                args = args[: args.index("args")]
            if len(args) != len(defs):
                raise NotImplementedError(
                    f"Class {name} must have default arguments for __init__ variables {init_fn.__code__.co_varnames[: -(len(init_fn.__code__.co_varnames - 1 - len(init_fn.__defaults__)))]}"
                )
        return super().__new__(cls, name, bases, dct)


class configBase(ABC, metaclass=configMeta):
    """
    Base abstract class that represents the normalisation/background settings for a scan.

    Parameters
    ----------
    scan : scan_base
        The scan number to apply the normalisation settings to.
    apply_to : list[str | int] | None, optional
        The list of scan channel names or scan channel indices to apply the normalisation to.
        If None, apply to all channels (default).
        Names are preferred over indices to avoid ambiguity when the scan structure changes.
    """

    def __init__(self, apply_to: list[str | int] | None = None) -> None:
        # Internal variables
        self._apply_to: list[str | int] | None = apply_to
        """
        The list of scan scan channel names (or indexes) to apply the normalisation to.
        If None, apply to all channels.
        """

    @property
    @abstractmethod
    def is_valid(self) -> bool:
        """
        Check if the normalisation settings are valid, i.e. data paths exist.

        Requires implementation in subclasses.

        Returns
        -------
        bool
            True if the settings are valid, False otherwise.
        """
        pass

    @property
    def apply_to(self) -> list[str | int] | None:
        """
        The list of scan channel names to apply the normalisation to.

        None implies apply normalisation to all channels.

        Parameters
        ----------
        apply_to : list[str] | None
            The list of scan channel names to apply the normalisation to.

        Returns
        -------
        list[str] | None
            The list of scan channel names to apply the normalisation to.
        """
        return self._apply_to

    @apply_to.setter
    def apply_to(self, apply_to: list[str | int] | None):
        self._apply_to = apply_to

    @apply_to.deleter
    def apply_to(self):
        self._apply_to = None


class normMethod(Enum):
    """
    Different methods of performing channel normalisation.

    Allows the user to specify how the normalisation should be performed.
    Values are stored as an enumerate.

    Attributes
    ----------
    NONE : 0
        No normalisation of the scan channel by the background channel.
    SUB : 1
        Common for background signals.
        Subtract the background channel from the scan channel.
    DIV : 2
        Common for flux signals. Divide the scan channel by the normalized
        (maximum value=1) background channel amplitude.
    MULT : 3
    """

    NONE = 0
    SUB = 1
    DIV = 2
    MULT = 3


class extSelection(Enum):
    """
    Definitions for choosing external data selection.

    Allow the user to describe how the data selection should be accessed.
    Values are stored as an enumerate.

    Attributes
    ----------
    NONE : 0
        No background normalisation of the scan.
    MOST_RECENT_PRIOR_KEYWORD : 1
        Use the most recent scan containing a keyword prior to the existing scan.
    CLOSEST_TIME_KEYWORD : 2
        Use the closest scan in time containing a keyword to the existing scan.
    MOST_RECENT_POST_KEYWORD : 3
        Use the most recent scan containing a keyword after the existing scan.
    FIXED_SCAN : 4
        Use a fixed scan object for background normalisation.
    """

    NONE = 0
    MOST_RECENT_PRIOR_KEYWORD = 1
    CLOSEST_TIME_KEYWORD = 2
    MOST_RECENT_POST_KEYWORD = 3
    FIXED_SCAN = 4


class configChannel(configBase):
    """
    Configurations where a particular channel is required to perform the normalisation.

    Parameters
    ----------
    norm_method : normMethod, optional
        The method of normalisation, by default normMethod.NONE.
    channel_name : str | None, optional
        The name of the channel to apply the normalisation to, by default None.
    apply_to : list[str] | None, optional
        The list of scan channel names to apply the normalisation to, by default None.
    """

    # Add the normMethod Enum as a class attribute
    normMethod = normMethod

    def __init__(
        self,
        norm_method: normMethod = normMethod.NONE,
        channel_name: str | None = None,
        apply_to: list[str | int] | None = None,
    ) -> None:
        # Initialise the normConfig class
        super().__init__(apply_to)

        # Initialise the internal variables
        self._norm_method: configChannel.normMethod = norm_method
        """The normalisation method choice."""
        self._channel_name: str | None = channel_name
        """The normalisation channel name choice."""

    @property
    def method(self) -> normMethod:
        """
        Property to get/set  the normalisation method.

        Can be one of the following:
        - norm_method.NONE
            No normalisation of the scan channel by the background channel.
        - norm_method.BACKGROUND
            Subtract the background channel from the scan channel.
        - norm_method.FLUX
            Divide the scan channel by the normalized (maximum value=1) background channel amplitude.

        Parameters
        ----------
        method : normMethod
            The normalisation method.

        Returns
        -------
        normMethod
            The normalisation method.
        """
        return self._norm_method

    @method.setter
    def method(self, method: normMethod):
        self._norm_method = method
        if method is self.normMethod.NONE:
            self._keyword = None
            self._path = None

    @property
    def channel(self) -> str | None:
        """
        The normalisation channel name.

        Parameters
        ----------
        channel_name : str | None
            The normalisation channel name.

        Returns
        -------
        str | None
            The normalisation channel name.
        """
        return self._channel_name

    @channel.setter
    def channel(self, channel_name: str | None):
        self._channel_name = channel_name

    @channel.deleter
    def channel(self) -> None:
        self._channel_name = None

    @configBase.is_valid.getter
    def is_valid(self) -> bool:
        """
        Check if the normalisation settings are valid, i.e. data paths exist.

        Returns
        -------
        bool
            True if the settings are valid, False otherwise.
        """
        return self._channel_name is not None or self.method is self.normMethod.NONE


class configExternal(configBase, ABC):
    """
    Abstract class for external data normalisation settings.

    Allows user to specify temporal proximity and keyword matching for external data.

    Parameters
    ----------
    channel_selection : extSelection, optional
        The method of selecting the external data, by default extSelection.NONE.
    keyword : str | None, optional
        The keyword to use for selecting the external data, by default None.
    path : str | None, optional
        The path to the external data, by default None.
    apply_to : list[str] | None, optional
        The list of scan channel names to apply the normalisation to, by default None.
    """

    extSelection = extSelection

    def __init__(
        self,
        channel_selection: extSelection = extSelection.NONE,
        keyword: str | None = None,
        path: str | None = None,
        apply_to: list[str | int] | None = None,
    ):
        # Initialise the normConfig class
        super().__init__(apply_to)
        # Initialise the internal variables
        self._channel_selection: configExternal.extSelection = channel_selection
        """The external data selection method."""
        self._keyword: str | None = keyword
        """The keyword selection used for various CHANNEL_SELECTION enumerate options."""
        self._path: str | None = path
        """The path selection corresponding to a file for the CHANNEL_SELECTION.FIXED_SCAN enumerate,
        or a directory for the CHANNEL_SELECTION.CLOSEST_TIME_KEYWORD enumerate."""

    @property
    def selection(self) -> extSelection:
        return self._channel_selection

    @selection.setter
    def selection(self, selection: extSelection):
        self._channel_selection = selection
        if (
            selection is self.extSelection.NONE
            or selection is self.extSelection.FIXED_SCAN
        ):
            self._keyword = None

    @property
    def keyword(self) -> str | None:
        return self._keyword

    @keyword.setter
    def keyword(self, keyword: str | None):
        self._keyword = keyword

    @property
    def path(self) -> str | None:
        return self._path

    @path.setter
    def path(self, path: str | None):
        if os.path.exists(path):
            self._path = path
            if os.path.isfile(path):
                # Automatically set selection to FIXED_SCAN if a file path is given instead of a directory.
                self._channel_selection = self.extSelection.FIXED_SCAN
        else:
            raise FileNotFoundError(f"Path `{path}` does not exist.")


class configExternalChannel(configChannel, configExternal):
    """
    Configurations for a single channel normalisation using external data.

    Parameters
    ----------
    channel_selection : extSelection, optional
        The method of selecting the external data, by default extSelection.NONE.
    keyword : str | None, optional
        The keyword to use for selecting the external data, by default None.
    path : str | None, optional
        The path to the external data, by default None.
    norm_method : normMethod, optional
        The method of normalisation, by default normMethod.NONE.
    channel_name : str | None, optional
        The name of the channel to apply the normalisation to, by default None.
    apply_to : list[str] | None, optional
        The list of scan channel names to apply the normalisation to, by default None.
    """

    # Add enumerates as class attributes
    extSelection = configExternal.extSelection
    normMethod = normMethod

    def __init__(
        self,
        channel_selection: configExternal.extSelection = extSelection.NONE,
        keyword: str | None = None,
        path: str | None = None,
        norm_method: normMethod = normMethod.NONE,
        channel_name: str | None = None,
        apply_to: list[str | int] | None = None,
    ) -> None:
        # Initialise the normConfigChannel class
        configChannel.__init__(self, norm_method, channel_name, apply_to=apply_to)
        configExternal.__init__(
            self, channel_selection, keyword, path, apply_to=apply_to
        )

    @configBase.is_valid.getter
    def is_valid(self) -> bool:
        """
        Check if the normalisation settings are valid, i.e. data paths exist.

        Returns
        -------
        bool
            True if the settings are valid, False otherwise.
        """
        match self.selection:
            case (
                self.extSelection.MOST_RECENT_PRIOR_KEYWORD
                | self.extSelection.CLOSEST_TIME_KEYWORD
                | self.extSelection.MOST_RECENT_POST_KEYWORD
            ):
                return (
                    self._keyword is not None
                    and self._path is not None
                    and os.path.exists(self._path)
                )
            case self.extSelection.FIXED_SCAN:
                return self._path is not None and os.path.exists(self._path)
            case self.extSelection.NONE:
                return True
            case _:
                return False


class configDouble(configBase):
    def __init__(self, apply_to: list[str] | None = None) -> None:
        super().__init__(apply_to)


# class configExternalData(configBase):
#     """
#     Configurations for external data normalisation.

#     Includes the method of normalisation and the source of the external data.
#     Does not include the data itself, when saving settings.

#     Parameters
#     ----------
#     norm_method : normMethod, optional
#         The method of normalisation, by default normMethod.NONE.
#     norm_source : normSource, optional
#         The source of the normalisation data, by default normSource.NONE.
#     handling_method : handlingMethod, optional
#         The method of handling the normalisation data, by default handlingMethod.NONE.
#     data_path : str, optional
#         The path to the normalisation data, by default None.
#     apply_to : list[str] | None, optional
#         The list of scan channel names to apply the normalisation to, by default None.
#     """

#     # Add the normMethod Enum as a class attribute to emulate data.
#     normMethod = normMethod

#     class normSource(Enum):
#         """
#         Enumerate flags to define the source of the external data.

#         Attributes
#         ----------
#         NONE : int
#             No external data source.
#         FILE : int
#             External data source is a file.
#         ARRAY : int
#             External data source is an array.
#         """

#         NONE = 0
#         FILE = 1
#         ARRAY = 2

#     class handlingMethod(Enum):
#         """
#         Enumerate flags to define how to apply external data to the scan.

#         Attributes
#         ----------
#         NONE : int
#             No data handling. Data must match the scan shape.
#         INTERPOLATE : int
#             Linearly interpolate the external data to match the scan shape.
#         SPLINE : int
#             Use a spline interpolation to match the external data to the scan shape.
#         """

#         NONE = 0
#         INTERPOLATE = 1
#         SPLINE = 2

#     def __init__(
#         self,
#         norm_method: normMethod = normMethod.NONE,
#         norm_source: normSource = normSource.NONE,
#         handling_method: handlingMethod = handlingMethod.NONE,
#         data_path: str | None = None,
#         apply_to: list[str] | None = None,
#     ) -> None:
#         self._norm_method: configExternalData.normMethod = norm_method
#         """Internal variable to track the normalisation method."""
#         self._norm_source: configExternalData.normSource = norm_source
#         """Internal variable to track the normalisation data source."""
#         self._handling_method: configExternalData.handlingMethod = handling_method
#         """Internal variable to track the normalisation data handling method."""
#         self._data_path: str | None = data_path
#         """Internal variable to track the normalisation data path."""
#         # Initialise the configBase class
#         super().__init__(apply_to)

#     @property
#     def method(self) -> normMethod:
#         """
#         A property defining the normalisation method.

#         Can be one of the following:
#         - DATA_METHOD.NONE
#             No normalisation performed.
#         - DATA_METHOD.BACKGROUND
#             Subtract the background data from the scan data.
#         - DATA_METHOD.FLUX
#             Divide the scan data by the normalized (maximum value=1) background data amplitude.

#         Returns
#         -------
#         normMethod
#             The normalisation method.
#         """
#         return self._norm_method

#     @method.setter
#     def method(self, method: normMethod) -> None:
#         self._norm_method = method

#     @property
#     def handling(self) -> handlingMethod:
#         """
#         A property defining the normalisation data handling method.

#         Parameters
#         ----------
#         handling_method : handlingMethod
#             The normalisation data handling method. Can be one of the following:
#             - HANDLING_METHOD.NONE
#                 No data handling. Data must match the scan shape.
#             - HANDLING_METHOD.INTERPOLATE
#                 Linearly interpolate the external data to match the scan shape.
#             - HANDLING_METHOD.SPLINE
#                 Use a spline interpolation to match the external data to the scan shape.

#         Returns
#         -------
#         handlingMethod
#             The normalisation data handling method.
#         """
#         return self._handling_method

#     @handling.setter
#     def handling(self, handling_method: handlingMethod) -> None:
#         self._handling_method = handling_method

#     @property
#     def source(self) -> normSource:
#         """
#         A property defining the normalisation data source.

#         Parameters
#         ----------
#         source : normSource
#             The normalisation data source. Can be one of the following:
#             - SOURCE.NONE
#                 No external data source.
#             - SOURCE.FILE
#                 External data source is a file.
#             - SOURCE.ARRAY
#                 External data source is an array.

#         Returns
#         -------
#         normSource
#             The normalisation data source.
#         """
#         return self._norm_source


# class configBackground(configBase):
#     """
#     Configurations for using a background scan for normalisation.
#     """

#     # TODO: Implement this class.


class edgeNormPre(Enum):
    """
    Enumerated definitions for pre-edge normalisation.

    Attributes
    ----------
    NONE : int
        No pre-edge normalisation.
    CONSTANT : int
        Constant normalisation over the edge domain.
    LINEAR : int
        Linear normalisation over the edge domain.
    EXPONENTIAL : int
        EXPONENTIAL normalisation over the edge domain.
    """

    NONE = 0
    CONSTANT = 1
    LINEAR = 2
    EXPONENTIAL = 3


class edgeNormPost(Enum):
    """
    Enumerated definitions for post-edge normalisation.

    Attributes
    ----------
    NONE : int
        No post-edge normalisation.
    CONSTANT : int
        Constant normalisation over the edge domain.
    """

    NONE = 0
    CONSTANT = 1


class configEdges(configBase):
    """
    A class to handle settings for edge normalisation.

    Parameters
    ----------
    pre_edge_norm_method : scan_normalised_edges.PREEDGE_NORM_TYPE, optional
        Method to normalise the pre-edge region, by default scan_normalised_edges.PREEDGE_NORM_TYPE.NONE.
    post_edge_norm_method : scan_normalised_edges.POSTEDGE_NORM_TYPE, optional
        Method to normalise the post-edge region, by default scan_normalised_edges.POSTEDGE_NORM_TYPE.NONE.
    pre_edge_domain : tuple[float, float] | None, optional
        Domain of the pre-edge region to normalise, by default None.
    post_edge_domain : tuple[float, float] | None, optional
        Domain of the post-edge region to normalise, by default None.
    pre_edge_level : float | None, optional
        Baseline level of the pre-edge region, by default None.
    post_edge_level : float | None, optional
        Baseline level of the post-edge region, by default None.
    apply_to : list[str] | None, optional
        The list of scan channel names to apply the normalisation to, by default None.

    Examples
    --------
    An example to normalise at the pre-edge and post-edge of sulfur K-edge.
    >>> edge_config = norm_config_edges(pre_edge_norm_method=scan_normalised_edges.PREEDGE_NORM_TYPE.LINEAR,
    ...                                 post_edge_norm_method=scan_normalised_edges.POSTEDGE_NORM_TYPE.NONE,
    ...                                 pre_edge_domain=(2460, 2465),
    ...                                 post_edge_domain=(2500, 2510),
    ...                                 pre_edge_level=0.0,
    ...                                 post_edge_level=1.0)
    """

    PREEDGE_NORM_TYPE = edgeNormPre
    """The pre-edge normalisation type enumerate."""
    POSTEDGE_NORM_TYPE = edgeNormPost
    """The post-edge normalisation type enumerate."""

    DEFAULT_PRE_EDGE_LEVEL_CONSTANT = 0.0
    """The constant offset for the pre-edge normalisation. I.e. Pre-edge at 0.0"""
    DEFAULT_PRE_EDGE_LEVEL_LINEAR = 0.0
    """The constant gradient for the pre-edge normalisation. I.e. Linear gradient at 0.0"""
    DEFAULT_PRE_EDGE_LEVEL_EXP = 0.0
    """The constant exponential level for the pre-edge normalisation. I.e. Exponential level at 0.0"""
    DEFAULT_POST_EDGE_LEVEL_CONSTANT = 1.0
    """The constant offset for the post-edge normalisation. I.e. Post-edge at 1.0"""

    def __init__(
        self,
        pre_edge_norm_method: PREEDGE_NORM_TYPE = PREEDGE_NORM_TYPE.NONE,
        post_edge_norm_method: POSTEDGE_NORM_TYPE = POSTEDGE_NORM_TYPE.NONE,
        pre_edge_domain: tuple[float, float] | list[int] | None = None,
        post_edge_domain: tuple[float, float] | list[int] | None = None,
        pre_edge_level: float | None = None,
        post_edge_level: float | None = None,
        apply_to: list[str] | None = None,
    ):
        super().__init__(apply_to)
        # Initialise some internal variables
        self._pre_edge_level = None
        self._post_edge_level = None
        self.pre_edge_norm_method = pre_edge_norm_method
        self.post_edge_norm_method = post_edge_norm_method
        self.pre_edge_domain = pre_edge_domain
        self.post_edge_domain = post_edge_domain

        # Set default values if not defined
        if pre_edge_level is None:
            self.pre_edge_level = self.DEFAULT_PRE_EDGE_LEVEL_CONSTANT
        else:
            self.pre_edge_level = pre_edge_level
        if post_edge_level is None:
            self.post_edge_level = self.DEFAULT_PRE_EDGE_LEVEL_CONSTANT
        else:
            self.post_edge_level = post_edge_level

    # Validation
    @configBase.is_valid.getter
    def is_valid(self) -> bool:
        """
        Check if the normalisation settings are valid.

        The minimum requirements for an edge normalisation are the pre-edge and post-edge domains.

        Returns
        -------
        bool
            True if the settings are valid, False otherwise.
        """
        # TODO more intelligent validation.
        return (
            self._pre_edge_domain is not None  # A domain exists
            and self._post_edge_domain is not None  # A domain exists
            # Check the domains are valid
            and (
                (
                    isinstance(self._pre_edge_domain, tuple)
                    and self._pre_edge_domain[0] < self._pre_edge_domain[1]
                )
                or not isinstance(self._pre_edge_domain, tuple)
            )
            and (
                (
                    isinstance(self._post_edge_domain, tuple)
                    and self._post_edge_domain[0] < self._post_edge_domain[1]
                )
                or not isinstance(self._post_edge_domain, tuple)
            )
        )

    # Edge Normalisation Method
    @property
    def pre_edge_norm_method(self) -> PREEDGE_NORM_TYPE:
        return self._pre_edge_norm_method

    @pre_edge_norm_method.setter
    def pre_edge_norm_method(self, method: PREEDGE_NORM_TYPE):
        # Check if method is valid
        if method in configEdges.PREEDGE_NORM_TYPE:
            self._pre_edge_norm_method = method
        else:
            raise ValueError(f"{method} not in {configEdges.PREEDGE_NORM_TYPE}.")
        # If pre-edge not defined, set to default level
        if self.pre_edge_level is None:
            match method:
                case configEdges.PREEDGE_NORM_TYPE.CONSTANT:
                    self.pre_edge_level = self.DEFAULT_PRE_EDGE_LEVEL_CONSTANT
                case configEdges.PREEDGE_NORM_TYPE.LINEAR:
                    self.pre_edge_level = self.DEFAULT_PRE_EDGE_LEVEL_LINEAR
                case configEdges.PREEDGE_NORM_TYPE.EXPONENTIAL:
                    self.pre_edge_level = self.DEFAULT_PRE_EDGE_LEVEL_EXP
                case _:
                    # Do nothing if not defined / NONE.
                    pass

        # Change pre-edge level by default to a reasonable value if setting exponential.
        if (
            method is configEdges.PREEDGE_NORM_TYPE.EXPONENTIAL
            and self.pre_edge_level <= 0
        ):
            self._pre_edge_level = self.DEFAULT_PRE_EDGE_LEVEL_EXP

        self._pre_edge_norm_method = method

    @pre_edge_norm_method.deleter
    def pre_edge_norm_method(self) -> None:
        self._pre_edge_norm_method = configEdges.PREEDGE_NORM_TYPE.NONE

    @property
    def post_edge_norm_method(self) -> POSTEDGE_NORM_TYPE:
        """
        Property to define the type of normalisation performed on the post-edge.

        Parameters
        ----------
        vals : normConfigEdges.EDGE_NORM_TYPE
            LINEAR, EXPONENTIAL or NONE.

        Returns
        -------
        normConfigEdges.EDGE_NORM_TYPE
            The current normalisation type.
        """
        return self._post_edge_norm_method

    @post_edge_norm_method.setter
    def post_edge_norm_method(self, method: POSTEDGE_NORM_TYPE):
        if method in configEdges.POSTEDGE_NORM_TYPE:
            self._post_edge_norm_method = method
        else:
            raise ValueError(f"{method} not in {configEdges.POSTEDGE_NORM_TYPE}.")

        # Set default if None
        if self._post_edge_level is None:
            match method:
                case configEdges.POSTEDGE_NORM_TYPE.CONSTANT:
                    self._post_edge_level = configEdges.DEFAULT_POST_EDGE_LEVEL_CONSTANT
                case _:
                    # Do nothing if not defined / NONE.
                    pass

    @post_edge_norm_method.deleter
    def post_edge_norm_method(self) -> None:
        self._post_edge_norm_method = configEdges.POSTEDGE_NORM_TYPE.NONE

    # Edge Normalisation Domain
    @property
    def pre_edge_domain(self) -> tuple[float, float] | list[int] | None:
        """
        A property defining the pre-edge domain of normalisation.

        If setting and `pre_edge_norm_method` is `NONE`, will set new normalisation enumerate to `LINEAR`.

        Parameters
        ----------
        vals : list[int] | tuple[float, float] | None
            Can be defined using a list of included indices matching x datapoints.
            Alternatively a tuple of two floats defining the inclusive endpoints.
            Alternatively None stops normalisation being performed on the pre-edge.

        Returns
        -------
        list[int] | tuple[float, float] | None
            Same as vals parameter.
        """
        return self._pre_edge_domain

    @pre_edge_domain.setter
    def pre_edge_domain(self, domain: tuple[float, float] | list[int] | None):
        if isinstance(domain, list):
            self._pre_edge_domain = domain.copy()  # non-immutable
        elif isinstance(domain, tuple) and len(domain) == 2:
            if domain[0] > domain[1]:
                raise ValueError(
                    f"Pre-edge domain contains lower-bound `{domain[0]}` which is greater than upperbound `{domain[1]}`."
                )
            self._pre_edge_domain = domain  # immutable, can't modify.
        elif domain is None:
            # Remove vals
            del self.pre_edge_domain
        else:
            raise ValueError(
                "The pre-edge domain needs to be defined by a list of integer indices, a tuple of inclusive endpoints or None."
            )
        # Default normalisation to linear if not already defined.
        if (
            self._pre_edge_norm_method is configEdges.PREEDGE_NORM_TYPE.NONE
            and domain is not None
        ):
            self._pre_edge_norm_method = configEdges.PREEDGE_NORM_TYPE.LINEAR

    @pre_edge_domain.deleter
    def pre_edge_domain(self):
        self._post_edge_domain = None

    @property
    def post_edge_domain(self) -> list[int] | tuple[float, float] | None:
        """
        A property defining the post-edge domain of normalisation.

        If setting and `post__apply_normalisation` is `NONE`, will set new normalisation enumerate to `LINEAR`.

        Parameters
        ----------
        vals : list[int] | tuple[float, float] | None
            Can be defined using a list of included indices matching x datapoints.
            Alternatively a tuple of two floats defining the inclusive endpoints.
            Alternatively None stops normalisation being performed on the post-edge.

        Returns
        -------
        list[int] | tuple[float, float] | None
            Same as vals parameter.
        """

        return self._post_edge_domain

    @post_edge_domain.setter
    def post_edge_domain(self, vals: list[int] | tuple[float, float] | None) -> None:
        if isinstance(vals, list):
            self._post_edge_domain = vals.copy()  # non-immutable
        elif isinstance(vals, tuple) and len(vals) == 2:
            if vals[0] > vals[1]:
                raise ValueError(
                    f"Post-edge domain contains lower-bound `{vals[0]}` which is greater than upperbound `{vals[1]}`."
                )
            self._post_edge_domain = vals  # immutable, can't modify.
        elif vals is None:
            # Remove vals and perform any other
            del self.post_edge_domain
        else:
            raise ValueError(
                "The post-edge domain needs to be defined by a list of integer indices, a tuple of inclusive endpoints or None."
            )
        # Default normalisation to constant if not already defined.
        if (
            self._post_edge_norm_method is configEdges.POSTEDGE_NORM_TYPE.NONE
            and vals is not None
        ):
            self._post_edge_norm_method = configEdges.POSTEDGE_NORM_TYPE.CONSTANT

    @post_edge_domain.deleter
    def post_edge_domain(self):
        self._post_edge_domain = None

    # Edge Normalisation Level
    @property
    def pre_edge_level(self) -> float:
        """
        Property to define the normalisation level (constant) for the pre-edge.

        Parameters
        ----------
        vals : float
            The normalisation average.

        Returns
        -------
        float
            The normalisation level.
        """
        return self._pre_edge_level

    @pre_edge_level.setter
    def pre_edge_level(self, level: float):
        if self.pre_edge_norm_method is configEdges.PREEDGE_NORM_TYPE.EXPONENTIAL:
            if level <= 0:
                raise ValueError(
                    "Exponential normalisation requires a positive, non-zero level."
                )
        self._pre_edge_level = level

    @pre_edge_level.deleter
    def pre_edge_level(self):
        if self.pre_edge_norm_method is configEdges.PREEDGE_NORM_TYPE.EXPONENTIAL:
            self._pre_edge_level = configEdges.DEFAULT_PRE_EDGE_LEVEL_EXP
        # elif self.pre_edge_norm_method is normConfigEdges.EDGE_NORM_TYPE.LINEAR:
        # self._pre_edge_level = normConfigEdges.DEFAULT_PRE_EDGE_LEVEL_LINEAR
        else:
            self._pre_edge_level = configEdges.DEFAULT_PRE_EDGE_LEVEL_LINEAR

    @property
    def post_edge_level(self) -> float:
        return self._post_edge_level

    @post_edge_level.setter
    def post_edge_level(self, level: float):
        self._post_edge_level = level

    @post_edge_level.deleter
    def post_edge_level(self):
        self._post_edge_level = configEdges.DEFAULT_PRE_EDGE_LEVEL_LINEAR


class configSeries(metaclass=configMeta):
    """
    Base class for a series of sequential normalisation configurations.

    Parameters
    ----------
    configs : list[normConfig] | None, optional
        The sequential list of normalisation configurations.
    """

    def __init__(self, configs: list[configBase] | None = None) -> None:
        if configs is None:
            self._configs: list[configBase] = []
        else:
            self._configs = configs

    def __len__(self):
        """
        Get the number of normalisation settings in the series.

        Returns
        -------
        int
            The number of normalisation settings in the series.
        """
        return self._configs.__len__()

    def __getitem__(self, index: int) -> configBase:
        """
        Get the normalisation settings at the specified index.

        Parameters
        ----------
        index : int
            The index of the normalisation settings to get.

        Returns
        -------
        configBase
            The normalisation settings at the specified index.
        """
        return self._configs[index]

    def __setitem__(self, index: int, value: configBase) -> None:
        """
        Set the normalisation settings at the specified index.

        Parameters
        ----------
        index : int
            The index of the normalisation settings to set.
        value : configBase
            The normalisation settings to set at the specified index.
        """
        self._configs[index] = value

    def yaml(self) -> str:
        """
        Convert the normalisation settings to yaml format.

        Returns
        -------
        str
            The yaml formatted string.
        """
        return yaml.dump(self._configs)

    def save(self, path: str, override: bool = False) -> None:
        """
        Save the normalisation settings to a yaml file.

        Parameters
        ----------
        path : str
            The path to save the settings file.
        override : bool, optional
            If True, override the file if it already exists, by default False.

        Raises
        ------
        FileExistsError
            If the file already exists and override is False.
        """
        if not path.endswith(".yaml") or not path.endswith(".yml"):
            path += ".yml"

        if not override and os.path.exists(path):
            raise FileExistsError(f"File {path} already exists.")

        with open(path, "w") as file:
            file.write(self.yaml())

    @staticmethod
    def loadstream(stream: str | TextIOWrapper) -> "configSeries":
        """
        Import a list of sequential normalisation settings from a yaml stream.

        Parameters
        ----------
        stream : str
            The yaml formatted string.

        Returns
        -------
        configSeries
            The loaded series of normalisation settings.
        """
        # Get list of normConfig class + subclasses
        config_clss: list[type[configBase]] = [configBase] + configBase.__subclasses__()
        # Get list of class attributes for settings
        attr_clss: list[type] = [
            val
            for cls in config_clss
            for val in cls.__dict__.values()
            if isinstance(val, type)
        ]
        # Use the union to define new objects
        clss = config_clss + attr_clss

        # Define the extra context for constructors
        def default_ctor(
            loader: yaml.SafeLoader, tag_suffix: str, node: yaml.MappingNode
        ):
            """
            Default constructor for unknown objects.

            Parameters
            ----------
            loader : yaml.SafeLoader
                The yaml loader to use.
            tag_suffix : str
                The suffix of the tag to match the class name.
            node : yaml.MappingNode
                The yaml node to construct the object from.

            Returns
            -------
            configBase
                The constructed object of the matching class.
            """
            # Get the class name from the tag
            class_name = tag_suffix.split(".")[-1]
            # Find the class that matches the tag
            for cls in clss:
                if cls.__name__ == class_name:
                    # Check if an enumerate type
                    if isinstance(cls, EnumType) and len(node.value) == 1:
                        n = node.value[0]
                        if "int" in n.tag and hasattr(n, "value"):
                            return cls(int(n.value))
                    else:
                        new_obj = cls()
                        for key, value in loader.construct_mapping(node).items():
                            setattr(new_obj, key, value)
                        return new_obj

        # Add the constructor to the yaml loader
        yaml.SafeLoader.add_multi_constructor("", default_ctor)
        data: list[configBase] = yaml.load(stream, Loader=yaml.SafeLoader)
        return configSeries(data)

    @staticmethod
    def load(path: str) -> "configSeries":
        """
        Load the sequential list of normalisation settings from a yaml file.

        Parameters
        ----------
        path : str
            The path to load the settings file.

        Returns
        -------
        configSeries
            The loaded series of normalisation settings.
        """
        with open(path, "r") as file:
            series = configSeries.loadstream(file)
        return series



"""
Main entry point for the pyNexafs package.

Runs the GUI for the package if the dependencies are met.
Can be run from the command line with the command `python -m pyNexafs`.
Allowed arguments:
- `--help`: Show help message and exit.
- `[directory]`: Specify a directory to load NEXAFS files from.
- If no arguments are provided, the GUI will open pointed at the user folder '~'.
"""

import sys
import os
from importlib.metadata import version

if __name__ == "__main__":
    args = sys.argv[1:]  # Get command line arguments, excluding the script name
    # Check if the script is run with the `--help` argument
    if "--help" in args:
        print(
            "pyNexafs: A Python package for analyzing NEXAFS data.\n"
            "Usage:\n"
            "\tpython -m pyNexafs [--help]\n"
            "\tpython -m pyNexafs [directory]\n"
            "Options:\n"
            "\t--help\tShow this help message and exit."
        )
        exit()

    if len(args) > 1:
        print("Error: Too many arguments. Use --help for usage information.")
        exit()

    dir_arg = args[0] if len(args) == 1 else None
    dir_exists = os.path.isdir(dir_arg) if dir_arg else False

    # Check if the dependencies are met
    try:
        import PyQt6

        assert PyQt6 is not None
        print(f"PyQt6 ({version('PyQt6')}) is installed. Starting the GUI...")
    except ImportError as e:
        print(
            "PyQt6 is not installed. Please install PyQt6 to run the GUI."
            + f"\nTraceback: {e}"
        )
        exit()

    from pyNexafs.gui.data_browser import gui

    # Create and run the app
    if dir_arg and dir_exists:
        gui(directory=os.path.normpath(dir_arg))
    else:
        gui()
